/*
 * mem.c -- An HP-2000 Timeshared BASIC System tape manipulation program.
 *
 * Copyright (c) 2005, Scott G. Taylor
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * SCOTT G. TAYLOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of Scott G. Taylor shall not
 * be used in advertising or otherwise to promote the sale, use or other dealings
 * in this Software without prior written authorization from Scott G. Taylor.
 */

/*
 * n_malloc() and n_free() are replacement functions for malloc(2) and free(2).
 * These functions track allocated memory and provide a mechanism for monitoring
 * memory use and assist in locating memory leaks.  n_list() provides a dump of
 * the currently-allocated memory list.
 */

#define MYMALLOC

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "config.h"
#include "hptsb.h"
#include "decl.h"

/* Total number of memory allocations that can be tracked. */
#define MEMALLOCSIZE	10240
/* Number of significant characters of a file name to track. */
#define SOURCEFILELEN	32

/* memalloc is the memory allocation tracking table. */
struct {
	void		*p;
	uint32_t	len;
	uint8_t		sourcefile[SOURCEFILELEN];
	uint32_t	lineno;
} memalloc[MEMALLOCSIZE];

/* alloctop points to the last memalloc structure that has been used. */
static uint32_t	alloctop = 0;

/*
 * n_malloc() - Provide memory allocation with usage tracking.
 */

void
*n_malloc (int len, char * sourcefile, int lineno)
{
	void	*p;
	
	if ((p = (void *) malloc (len)) == NULL)
		eprintf("n_malloc: Can't malloc %d bytes for file %s, line %d: ",
			sourcefile, lineno, len, strerror(errno));
	else if (alloctop == MEMALLOCSIZE) {
		eprintf("n_malloc: Can't track allocation of %d bytes for file %s, line %d: "
				"Memory table full.   Increase MEMALLOCSIZE above %d.\n", 
			len, sourcefile, lineno, MEMALLOCSIZE);
	} else {
		dprintf (DEBUG_MEMORY, "n_malloc: Allocating 0x%x for %d bytes: file %s, line %d\n",
				p, len, sourcefile, lineno);
		memalloc[alloctop].p = p;
		memalloc[alloctop].lineno = lineno;
		strncpy(memalloc[alloctop].sourcefile, sourcefile, SOURCEFILELEN - 1);
		memalloc[alloctop].sourcefile[SOURCEFILELEN - 1] = 0;
		memalloc[alloctop].len = len;
		alloctop++;
	}
	return p;
}

/*
 * n_free() - Frees memory allocated by n_malloc().
 */

void
n_free (void * p, char * sourcefile, int lineno)
{
	uint32_t	i;
	if (p == NULL) {
		eprintf("n_free: free() of NULL pointer in sourcefile %s, lineno %d\n",
				sourcefile, lineno);
		return;
	}
	for (i = 0; i < alloctop && p != memalloc[i].p; i++);
	if (i == alloctop) {
		eprintf("n_free: Attempting to free unknown pointer 0x%x: %s (%d)\n",
				p, sourcefile, lineno);
		return;
	}
	dprintf (DEBUG_MEMORY, "n_malloc: Freeing 0x%x: file %s, line %d\n",
				p, sourcefile, lineno);
	alloctop--;
	memalloc[i].p = memalloc[alloctop].p;
	memalloc[i].len = memalloc[alloctop].len;
	strcpy(memalloc[i].sourcefile, memalloc[alloctop].sourcefile); 
	memalloc[i].lineno = memalloc[alloctop].lineno;
	free(p);
}

/*
 * n_list - Produce a listing of memory currently allocated through n_malloc().
 */

void n_list (uint32_t debug_flags)
{
	uint32_t	i;
	dprintf (debug_flags, "MEMORY TABLE:\n");
	for (i = 0; i < alloctop; i++)
		dprintf (debug_flags, "  0x%x	%d	%d	%-.15s\n",
			memalloc[i].p, memalloc[i].len, memalloc[i].lineno, memalloc[i].sourcefile);
}
